# -*- coding: utf-8 -*-

"""
Reegis geometry tools.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"


# Python libraries
import os
from collections import namedtuple

# Internal libraries
from deflex import config as cfg
from reegis import geometries as geo


def deflex_regions(rmap=None, rtype='polygons'):
    """

    Parameters
    ----------
    rmap : str
        Name of the deflex map.
    rtype : str
        Type of the deflex map ('polygon', 'labels').

    Returns
    -------
    GeoDataFrame

    Examples
    --------
    >>> regions=deflex_regions('de17')
    >>> len(regions)
    17
    >>> regions.geometry.iloc[0].geom_type
    'MultiPolygon'
    >>> l=deflex_regions('de21', 'labels').loc['DE04', 'geometry']
    >>> l.geom_type
    'Point'
    >>> l.x
    13.2
    >>> l.y
    51.1
    >>> cfg.tmp_set('init', 'map', 'de22')
    >>> deflex_regions().name
    'de22'
    >>> list(deflex_regions('de02').index)
    ['DE01', 'DE02']
    """
    if rmap is None:
        rmap = cfg.get("init", "map")
    name = os.path.join(
        cfg.get("paths", "geo_deflex"),
        cfg.get("geometry", "deflex_polygon").format(
            suffix=".geojson", map=rmap, type=rtype
        ),
    )
    regions = geo.load(fullname=name)
    regions.set_index("region", inplace=True)
    regions.name = rmap
    return regions


def deflex_power_lines(rmap=None, rtype="lines"):
    """

    Parameters
    ----------
    rmap : str
        Name of the deflex powerline map.
    rtype : str
        Type of the deflex powerline map ('lines', 'labels').

    Returns
    -------

    Examples
    --------
    >>> lines=deflex_power_lines('de17')
    >>> lines.geometry.iloc[0].geom_type
    'LineString'
    >>> len(lines)
    31
    >>> deflex_power_lines('de02').index[0]
    'DE01-DE02'
    >>> cfg.tmp_set('init', 'map', 'de21')
    >>> deflex_power_lines().name
    'de21'
    """
    if rmap is None:
        rmap = cfg.get("init", "map")
    name = os.path.join(
        cfg.get("paths", "geo_deflex"),
        cfg.get("geometry", "powerlines").format(
            map=rmap, type=rtype, suffix=".geojson"
        ),
    )
    lines = geo.load(fullname=name)
    lines.set_index("name", inplace=True)
    lines.name = rmap
    return lines


def divide_off_and_onshore(regions):
    """
    Sort regions into onshore and offshore regions. A namedtuple with two list
    of regions ids will be returned. Fetch the `onshore` and `offshore`
    attribute of the named tuple to get the list.

    Parameters
    ----------
    regions : GeoDataFrame
        A region set with the region id in the index.

    Returns
    -------
    named tuple

    Examples
    --------
    >>> reg=deflex_regions('de02')
    >>> divide_off_and_onshore(reg).onshore
    ['DE01']
    >>> reg=deflex_regions('de21')
    >>> divide_off_and_onshore(reg).offshore
    ['DE19', 'DE20', 'DE21']
    """
    region_type = namedtuple("RegionType", "offshore onshore")
    regions_centroid = regions.copy()
    regions_centroid.geometry = regions_centroid.centroid

    germany_onshore = geo.load(
        cfg.get("paths", "geometry"), cfg.get("geometry", "germany_polygon")
    )

    gdf = geo.spatial_join_with_buffer(
        regions_centroid, germany_onshore, "onshore", limit=0
    )

    onshore = list(gdf.loc[gdf.onshore == 0].index)
    offshore = list(gdf.loc[gdf.onshore == "unknown"].index)

    return region_type(offshore=offshore, onshore=onshore)



# -*- coding: utf-8 -*-

"""Processing a list of power plants in Germany.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"


# Python libraries
import os
import logging

# internal modules
from deflex import config as cfg
from reegis import demand_heat


def get_heat_profiles_deflex(deflex_geo, year, time_index=None,
                             weather_year=None, keep_unit=False):
    """

    Parameters
    ----------
    year
    deflex_geo
    time_index
    weather_year
    keep_unit

    Returns
    -------

    """
    # separate_regions=keep all demand connected to the region
    separate_regions = cfg.get_list("demand_heat", "separate_heat_regions")

    # add second fuel to first
    combine_fuels = cfg.get_dict("combine_heat_fuels")

    # fuels to be dissolved per region
    region_fuels = cfg.get_list("demand_heat", "local_fuels")

    fn = os.path.join(
        cfg.get("paths", "demand"),
        "heat_profiles_{year}_{map}".format(year=year, map=deflex_geo.name),
    )

    demand_region = (
        demand_heat.get_heat_profiles_by_region(
            deflex_geo, year, to_csv=fn, weather_year=weather_year
        )
        .groupby(level=[0, 1], axis=1)
        .sum()
    )

    # Decentralised demand is combined to a nation-wide demand if not part
    # of region_fuels.
    regions = list(
        set(demand_region.columns.get_level_values(0).unique())
        - set(separate_regions)
    )

    # If region_fuels is 'all' fetch all fuels to be local.
    if "all" in region_fuels:
        region_fuels = demand_region.columns.get_level_values(1).unique()

    for fuel in demand_region.columns.get_level_values(1).unique():
        demand_region["DE_demand", fuel] = 0

    for region in regions:
        for f1, f2 in combine_fuels.items():
            demand_region[region, f1] += demand_region[region, f2]
            demand_region.drop((region, f2), axis=1, inplace=True)
        cols = list(set(demand_region[region].columns) - set(region_fuels))
        for col in cols:
            demand_region["DE_demand", col] += demand_region[region, col]
            demand_region.drop((region, col), axis=1, inplace=True)

    if time_index is not None:
        demand_region.index = time_index

    if not keep_unit:
        msg = (
            "The unit of the source is 'TJ'. "
            "Will be divided by {0} to get 'MWh'."
        )
        converter = 0.0036
        demand_region = demand_region.div(converter)
        logging.debug(msg.format(converter))

    demand_region.sort_index(1, inplace=True)

    for c in demand_region.columns:
        if demand_region[c].sum() == 0:
            demand_region.drop(c, axis=1, inplace=True)

    return demand_region


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Main script.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"


# Python libraries
import os
import logging
from datetime import datetime
import time
import traceback

# internal modules
from deflex import config as cfg
from deflex import scenario_tools


def stopwatch():
    """Track the running time."""
    if not hasattr(stopwatch, "start"):
        stopwatch.start = datetime.now()
    return str(datetime.now() - stopwatch.start)[:-7]


def main_secure(year, rmap, csv=True, es=None, plot_graph=False,
                extra_regions=None):
    """

    Parameters
    ----------
    ----------
    year : int
        Year of an existing  basic scenario.
    rmap : str
        A valid deflex map id (de02, de17, de21, de22) of an existing scenario.
    csv : bool
        Use csv collection. If set to False the xls-file is used.
    es : oemof.solph.EnergySystem
        A valid energy system if needed.
    plot_graph : bool
        Set to True to plot the energy system.
    extra_regions : list
        Use separate resource buses for these regions.

    Returns
    -------

    Examples
    --------
    >>> main_secure(2014, 'de21')  # doctest: +SKIP
    """
    try:
        main(year, rmap, csv=csv, es=es, plot_graph=plot_graph,
             extra_regions=extra_regions)
    except Exception as e:
        logging.error(traceback.format_exc())
        time.sleep(0.5)
        logging.error(e)
        time.sleep(0.5)


def main(year, rmap, csv=True, es=None, plot_graph=False, extra_regions=None):
    """

    Parameters
    ----------
    year : int
        Year of an existing  basic scenario.
    rmap : str
        A valid deflex map id (de02, de17, de21, de22) of an existing scenario.
    csv : bool
        Use csv collection. If set to False the xls-file is used.
    es : oemof.solph.EnergySystem
        A valid energy system if needed.
    plot_graph : bool
        Set to True to plot the energy system.
    extra_regions : list
        Use separate resource buses for these regions.

    Returns
    -------

    Examples
    --------
    >>> main(2014, 'de21')  # doctest: +SKIP
    """
    stopwatch()
    cfg.tmp_set("init", "map", rmap)
    name = "{0}_{1}_{2}".format("deflex", year, cfg.get("init", "map"))

    path = os.path.join(cfg.get("paths", "scenario"), "deflex", str(year))

    if csv is True:
        csv_dir = name + "_csv"
        csv_path = os.path.join(path, csv_dir)
        excel_path = None
    else:
        excel_path = os.path.join(path, name + ".xls")
        csv_path = None

    model_scenario(
        xls_file=excel_path,
        csv_path=csv_path,
        res_path=path,
        name=name,
        rmap=rmap,
        year=year,
        es=es,
        plot_graph=plot_graph,
        extra_regions=extra_regions,
    )


def model_scenario(
    xls_file=None,
    csv_path=None,
    res_path=None,
    name="noname",
    rmap=None,
    year="unknown",
    es=None,
    plot_graph=False,
    extra_regions=None,
):
    """
    Compute a deflex scenario.

    Parameters
    ----------
    xls_file : str
        Full filename to a valid xls-file.
    csv_path : str
        Full path to a valid csv-collection.
    res_path : str
        Path to store the output file. If None the results will be stored along
        with the scenarios.
    name : str
        The name of the scenario.
    year : int
        The year or year-range of the scenario.
    rmap : str
        The name of the used region map.
    es : oemof.solph.EnergySystem
        A valid energy system if needed.
    plot_graph : bool
        Set to True to plot the energy system.
    extra_regions : list
        Use separate resource buses for these regions.

    Returns
    -------

    Examples
    --------
    >>> model_scenario('/my/path/to/scenario.xls', name='my_scenario',
    ...                rmap='deXX', year=2025)  # doctest: +SKIP
    """
    stopwatch()

    if xls_file is not None and csv_path is not None:
        raise ValueError("It is not allowed to define more than one input.")

    meta = {
        "year": year,
        "model_base": "deflex",
        "map": rmap,
        "solver": cfg.get("general", "solver"),
        "start_time": datetime.now(),
    }

    sc = scenario_tools.DeflexScenario(name=name, year=2014, meta=meta)
    if es is not None:
        sc.es = es

    if csv_path is not None:
        if res_path is None:
            res_path = os.path.dirname(csv_path)
        logging.info(
            "Read scenario from csv collection: {0}".format(stopwatch())
        )
        sc.load_csv(csv_path)
    elif xls_file is not None:
        if res_path is None:
            res_path = os.path.dirname(xls_file)
        logging.info("Read scenario from xls-file: {0}".format(stopwatch()))
        sc.load_excel(xls_file)

    if extra_regions is not None:
        sc.extra_regions = extra_regions

    logging.info("Add nodes to the EnergySystem: {0}".format(stopwatch()))
    sc.table2es()

    # Save energySystem to '.graphml' file if plot_graph is True
    if plot_graph:
        sc.plot_nodes(
            filename=os.path.join(res_path, name),
            remove_nodes_with_substrings=["bus_cs"],
        )

    logging.info("Create the concrete model: {0}".format(stopwatch()))
    sc.create_model()

    logging.info("Solve the optimisation model: {0}".format(stopwatch()))
    sc.solve(solver=cfg.get("general", "solver"))

    logging.info("Solved. Dump results: {0}".format(stopwatch()))
    res_path = os.path.join(
        res_path, "results_{0}".format(cfg.get("general", "solver"))
    )
    os.makedirs(res_path, exist_ok=True)
    out_file = os.path.join(res_path, name + ".esys")
    logging.info("Dump file to {0}".format(out_file))
    sc.meta["end_time"] = datetime.now()
    sc.dump_es(out_file)

    logging.info(
        "All done. deflex finished without errors: {0}".format(stopwatch())
    )


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Processing a list of power plants in Germany.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"


# Python libraries
import os

# External libraries
import pandas as pd
import math

# internal modules
from deflex import config as cfg

from deflex import geometries


def get_grid_capacity(grid, plus, minus):
    """Read the grid capacity from a given region pair from the renpass db."""
    tmp_grid = grid.query(
        "plus_region_id == {:0d} & ".format(plus)
        + "minus_region_id == {:1d} & ".format(minus)
        + "scenario_name == 'status_quo_2012_distance'"
    )

    if len(tmp_grid) > 0:
        capacity = tmp_grid.capacity_calc.sum()
        distance = tmp_grid.distance.iloc[0]
    else:
        capacity = 0
        distance = 0
    return capacity, distance


def add_reverse_direction(df):
    """
    Duplicate all entries of a DataFrame with a reverse index. The index must
    contain a dash between two sub-strings.
    """
    values = df.copy()

    def id_inverter(name):
        """Swap the sub-parts of a string left and right of a dash."""
        return "-".join([name.split("-")[1], name.split("-")[0]])

    df.index = df.index.map(id_inverter)

    return pd.DataFrame(pd.concat([values, df]))


def get_electrical_transmission_default(
    rmap=None, power_lines=None, both_directions=False
):
    """
    Creates a default set of transmission capacities, distance and efficiency.
    The map of the lines must exist in the geometries directory. The default
    values are infinity for the capacity, nan for the distance and 1 for the
    efficiency.

    Parameters
    ----------
    rmap : str
        The name of the transmission line map, that is part of deflex.
    power_lines : iterable[str]
        A list of names of transmission lines. All name must contain a dash
        between the id of the regions (FromRegion-ToRegion).
    both_directions : bool
        If True any line will be replicated in the reverse direction.

    Returns
    -------
    pd.DataFrame
        Transmission capacity, distance and efficiency between regions

    Examples
    --------
    >>> df=get_electrical_transmission_default('de21')
    >>> df.loc['DE10-DE12', 'capacity']
    inf
    >>> df.loc['DE10-DE12', 'distance']
    nan
    >>> df.loc['DE10-DE12', 'efficiency']
    1.0
    >>> len(df)
    39
    >>> len(get_electrical_transmission_default('de22'))
    40
    >>> len(get_electrical_transmission_default('de17'))
    31
    >>> len(get_electrical_transmission_default('de02'))
    1
    >>> my_lines=['reg1-reg2', 'reg2-reg3']
    >>> df=get_electrical_transmission_default(power_lines=my_lines)
    >>> df.loc['reg1-reg2', 'capacity']
    inf
    >>> df=get_electrical_transmission_default(power_lines=my_lines,
    ...                                          both_directions=True)
    >>> df.loc['reg2-reg1', 'capacity']
    inf

    """
    if power_lines is None:
        power_lines = pd.DataFrame(geometries.deflex_power_lines(rmap)).index

    df = pd.DataFrame()
    for l in power_lines:
        df.loc[l, "capacity"] = float("inf")
        df.loc[l, "distance"] = float("nan")
        df.loc[l, "efficiency"] = 1

    if both_directions is True:
        df = add_reverse_direction(df)
    return df


def get_electrical_transmission_renpass(both_directions=False):
    """
    Prepare the transmission capacity and distance between de21 regions from
    the renpass database. The original table of the reegis database is
    transferred to a csv file, which is part of the reegis package. As renpass
    is deprecated it will not change in the future. The index uses the format
    'region1-region2'. The distance is taken from centroid to centroid. By
    default every region pair exists only once. It is possible to get an entry
    in both directions if the parameter `both_directions` is set True.

    The capacity calculation is taken from the description of the renpass
    package [1]_. The data is taken from the renpass database [2]_.

    This function is only valid for the original renpass region set.

    Parameters
    ----------
    both_directions : bool
        If True any line will be replicated in the reverse direction.

    Returns
    -------
    pd.DataFrame
        Transmission capacity and distance between regions

    References
    ----------
    .. [1] Wiese, Frauke (2015). „Renewable Energy Pathways Simulation
        System – Open Source as an approach to meet challenges in energy
        modeling“. Diss. University of Flensburg. URL :
        https://www.reiner-lemoine-stiftung.de/pdf/dissertationen/Dissertation_Frauke_Wiese.pdf.
        (page 49)
    .. [2] Wiese, F.: Renpass - Renewable Energy Pathways Simulation System,
        https://github.com/fraukewiese/renpass

    Examples
    --------
    >>> df=get_electrical_transmission_renpass()
    >>> int(df.loc['DE11-DE17', 'capacity'])
    2506
    >>> int(df.loc['DE18-DE17', 'distance'])
    119
    >>> df.loc['DE08-DE06']
    capacity    7519.040402
    distance     257.000000
    Name: DE08-DE06, dtype: float64
    >>> df=get_electrical_transmission_renpass(both_directions=True)
    >>> int(df.loc['DE11-DE17', 'capacity'])
    2506
    >>> int(df.loc['DE17-DE11', 'capacity'])
    2506
    """
    f_security = cfg.get("transmission", "security_factor")
    current_max = cfg.get("transmission", "current_max")

    grid = pd.read_csv(
        os.path.join(
            cfg.get("paths", "data_deflex"),
            cfg.get("transmission", "transmission_renpass"),
        )
    )

    grid["capacity_calc"] = (
        grid.circuits
        * current_max
        * grid.voltage
        * f_security
        * math.sqrt(3)
        / 1000
    )

    pwr_lines = pd.DataFrame(geometries.deflex_power_lines())

    for l in pwr_lines.index:
        split = l.split("-")
        a = int("110{0}".format(split[0][2:]))
        b = int("110{0}".format(split[1][2:]))
        # print(a, b)
        cap1, dist1 = get_grid_capacity(grid, a, b)
        cap2, dist2 = get_grid_capacity(grid, b, a)

        if cap1 == 0 and cap2 == 0:
            pwr_lines.loc[l, "capacity"] = 0
            pwr_lines.loc[l, "distance"] = 0
        elif cap1 == 0:
            pwr_lines.loc[l, "capacity"] = cap2
            pwr_lines.loc[l, "distance"] = dist2
        elif cap2 == 0:
            pwr_lines.loc[l, "capacity"] = cap1
            pwr_lines.loc[l, "distance"] = dist1

    # plot_grid(pwr_lines)
    df = pwr_lines[["capacity", "distance"]]

    if both_directions is True:
        df = add_reverse_direction(df)

    return df


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Adapting the general reegis power plants to the deflex model.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"


import pandas as pd
import os
import logging
from reegis import geometries as reegis_geometries
from deflex import config as cfg
from reegis import powerplants


# Todo: Revise and test.


def pp_reegis2deflex(regions, name, filename_in=None, filename_out=None):
    """
    Add federal states and deflex regions to powerplant table from reegis. As
    the process takes a while the result is stored for further usage.

    Returns
    -------
    str : The full path where the result file is stored.

    """
    if filename_out is None:
        filename_out = os.path.join(
            cfg.get("paths", "powerplants"),
            cfg.get("powerplants", "deflex_pp"),
        ).format(map=cfg.get("init", "map"))

    # Add deflex regions to powerplants
    pp = powerplants.add_regions_to_powerplants(
        regions, name, dump=False, filename=filename_in
    )

    # Add federal states to powerplants
    federal_states = reegis_geometries.get_federal_states_polygon()
    pp = powerplants.add_regions_to_powerplants(
        federal_states, "federal_states", pp=pp, dump=False
    )

    # store the results for further usage of deflex
    pp.to_hdf(filename_out, "pp")
    return filename_out


# def remove_onshore_technology_from_offshore_regions(df):
#     """ This filter should be improved. It is slow and has to be adapted
#     manually. Anyhow it seems to work this way."""
#
#     logging.info("Removing onshore technology from offshore regions.")
#     logging.info("The code is not efficient. So it may take a while.")
#
#     offshore_regions=(
#         cfg.get_dict_list('offshore_regions_set')[cfg.get('init', 'map')])
#
#     coast_regions={'de02': {'MV': 'DE01',
#                               'SH': 'DE01',
#                               'NI': 'DE01 '},
#                      'de17': {'MV': 'DE13',
#                               'SH': 'DE01',
#                               'NI': 'DE03'},
#                      'de21': {'MV': 'DE01',
#                               'SH': 'DE13',
#                               'NI': 'DE14'},
#                      'de22': {'MV': 'DE01',
#                               'SH': 'DE13',
#                               'NI': 'DE14'}}
#     try:
#         dc=coast_regions[cfg.get('init', 'map')]
#     except KeyError:
#         raise ValueError('Coast regions not defined for {0} model.'.format(
#             cfg.get('init', 'map')))
#
#     region_column='{0}_region'.format(cfg.get('init', 'map'))
#
#     for ttype in ['Solar', 'Bioenergy', 'Wind']:
#         for region in offshore_regions:
#             logging.debug("Clean {1} from {0}.".format(region, ttype))
#
#             c1=df['energy_source_level_2'] == ttype
#             c2=df[region_column] == region
#
#             condition=c1 & c2
#
#             if ttype == 'Wind':
#                 condition=c1 & c2 & (df['technology'] == 'Onshore')
#
#             for i, v in df.loc[condition].iterrows():
#                 df.loc[i, region_column]=(
#                     dc[df.loc[i, 'federal_states']])
#     return df


def process_pp_table(pp):
    # # Remove powerplants outside Germany
    # for state in cfg.get_list('powerplants', 'remove_states'):
    #     pp=pp.loc[pp.state != state]
    #
    # if clean_offshore:
    #     pp=remove_onshore_technology_from_offshore_regions(pp)
    # Remove PHES (storages)
    if cfg.get("powerplants", "remove_phes"):
        pp = pp.loc[pp.technology != "Pumped storage"]
    return pp


def get_deflex_pp_by_year(regions, year, name, overwrite_capacity=False):
    """

    Parameters
    ----------
    regions : GeoDataFrame
    year : int
    name : str
    overwrite_capacity : bool
        By default (False) a new column "capacity_<year>" is created. If set to
        True the old capacity column will be overwritten.

    Returns
    -------

    """
    filename = os.path.join(
        cfg.get("paths", "powerplants"), cfg.get("powerplants", "deflex_pp")
    ).format(map=name)
    logging.info("Get deflex power plants for {0}.".format(year))
    if not os.path.isfile(filename):
        msg = "File '{0}' does not exist. Will create it from reegis file."
        logging.debug(msg.format(filename))
        filename = pp_reegis2deflex(regions, name)
    pp = pd.DataFrame(pd.read_hdf(filename, "pp", mode="r"))

    # Remove unwanted data sets
    pp = process_pp_table(pp)

    filter_columns = ["capacity_{0}", "capacity_in_{0}"]

    # Get all powerplants for the given year.
    # If com_month exist the power plants will be considered month-wise.
    # Otherwise the commission/decommission within the given year is not
    # considered.

    for fcol in filter_columns:
        filter_column = fcol.format(year)
        orig_column = fcol[:-4]
        c1 = (pp["com_year"] < year) & (pp["decom_year"] > year)
        pp.loc[c1, filter_column] = pp.loc[c1, orig_column]

        c2 = pp["com_year"] == year
        pp.loc[c2, filter_column] = (
            pp.loc[c2, orig_column] * (12 - pp.loc[c2, "com_month"]) / 12
        )
        c3 = pp["decom_year"] == year
        pp.loc[c3, filter_column] = (
            pp.loc[c3, orig_column] * pp.loc[c3, "com_month"] / 12
        )

        if overwrite_capacity:
            pp[orig_column] = 0
            pp[orig_column] = pp[filter_column]
            del pp[filter_column]

    return pp


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Work with the scenario data.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"

# Python libraries
import os
import sys
import calendar
import datetime
import shutil
import dill as pickle
from collections import namedtuple
import logging

# External libraries
import networkx as nx
import pandas as pd
from matplotlib import pyplot as plt

# oemof libraries
from oemof import solph
from oemof.tools import logger, helpers
from oemof import outputlib
from oemof import graph

if sys.getrecursionlimit() < 3000:
    sys.setrecursionlimit(3000)


class NodeDict(dict):
    __slots__ = ()

    def __setitem__(self, key, item):
        if super().get(key) is None:
            super().__setitem__(key, item)
        else:
            msg = (
                "Key '{0}' already exists. ".format(key)
                + "Duplicate keys are not allowed in a node dictionary."
            )
            raise KeyError(msg)


class Scenario:
    def __init__(self, **kwargs):
        self.name = kwargs.get("name", "unnamed_scenario")
        self.table_collection = kwargs.get("table_collection", {})
        self.year = kwargs.get("year", None)
        self.ignore_errors = kwargs.get("ignore_errors", False)
        self.round_values = kwargs.get("round_values", 0)
        self.model = kwargs.get("model", None)
        self.es = kwargs.get("es", None)
        self.results = None
        self.results_fn = kwargs.get("results_fn", None)
        self.debug = kwargs.get("debug", None)
        self.location = None
        self.map = None
        self.meta = kwargs.get("meta", None)

    def initialise_energy_system(self):
        if self.debug is True:
            number_of_time_steps = 3
        else:
            try:
                if calendar.isleap(self.year):
                    number_of_time_steps = 8784
                else:
                    number_of_time_steps = 8760
            except TypeError:
                msg = (
                    "You cannot create an EnergySystem with self.year={0}, "
                    "of type {1}."
                )
                raise TypeError(msg.format(self.year, type(self.year)))

        date_time_index = pd.date_range(
            "1/1/{0}".format(self.year), periods=number_of_time_steps, freq="H"
        )
        return solph.EnergySystem(timeindex=date_time_index)

    def load_excel(self, filename=None):
        """Load scenario from an excel-file."""
        if filename is not None:
            self.location = filename
        xls = pd.ExcelFile(self.location)
        for sheet in xls.sheet_names:
            self.table_collection[sheet] = xls.parse(
                sheet, index_col=[0], header=[0, 1]
            )
        return self

    def load_csv(self, path=None):
        """Load scenario from a csv-collection."""
        if path is not None:
            self.location = path
        for file in os.listdir(self.location):
            if file[-4:] == ".csv":
                filename = os.path.join(self.location, file)
                self.table_collection[file[:-4]] = pd.read_csv(
                    filename, index_col=[0], header=[0, 1]
                )
        return self

    def to_excel(self, filename):
        """Dump scenario into an excel-file."""
        # create path if it does not exist
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        writer = pd.ExcelWriter(filename)
        for name, df in sorted(self.table_collection.items()):
            df.to_excel(writer, name)
        writer.save()
        logging.info("Scenario saved as excel file to {0}".format(filename))

    def to_csv(self, path):
        """Dump scenario into a csv-collection."""
        if os.path.isdir(path):
            shutil.rmtree(os.path.join(path))
        os.makedirs(path)

        for name, df in self.table_collection.items():
            name = name.replace(" ", "_") + ".csv"
            filename = os.path.join(path, name)
            df.to_csv(filename)
        logging.info("Scenario saved as csv-collection to {0}".format(path))

    def check_table(self, table_name):
        if self.table_collection[table_name].isnull().values.any():
            c = []
            for column in self.table_collection[table_name].columns:
                if self.table_collection[table_name][column].isnull().any():
                    c.append(column)
            msg = "Nan Values in the {0} table (columns: {1})."
            raise ValueError(msg.format(table_name, c))
        return self

    def create_nodes(self):
        pass

    def initialise_es(self, year=None):
        if year is not None:
            self.year = year
        self.es = self.initialise_energy_system()
        return self

    def add_nodes(self, nodes):
        """

        Parameters
        ----------
        nodes : dict
            Dictionary with a unique key and values of type oemof.network.Node.

        Returns
        -------
        self

        """
        if self.es is None:
            self.initialise_es()
        self.es.add(*nodes.values())
        return self

    def table2es(self):
        if self.es is None:
            self.es = self.initialise_energy_system()
        nodes = self.create_nodes()
        self.es.add(*nodes.values())
        return self

    def create_model(self):
        self.model = solph.Model(self.es)
        return self

    def dump_es(self, filename):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        f = open(filename, "wb")
        if self.meta is None:
            if self.es.results is not None and "Meta" in self.es.results:
                self.meta = self.es.results["meta"]
        pickle.dump(self.meta, f)
        pickle.dump(self.es.__dict__, f)
        f.close()
        logging.info("Results dumped to {0}.".format(filename))

    def restore_es(self, filename=None):
        if filename is None:
            filename = self.results_fn
        else:
            self.results_fn = filename
        if self.es is None:
            self.es = solph.EnergySystem()
        f = open(filename, "rb")
        self.meta = pickle.load(f)
        self.es.__dict__ = pickle.load(f)
        f.close()
        self.results = self.es.results["main"]
        logging.info("Results restored from {0}.".format(filename))

    def scenario_info(self, solver_name):
        sc_info = {
            "name": self.name,
            "datetime": datetime.datetime.now(),
            "year": self.year,
            "solver": solver_name,
        }
        return sc_info

    def solve(self, with_duals=False, tee=True, logfile=None, solver=None):
        logging.info("Optimising using {0}.".format(solver))

        if with_duals:
            self.model.receive_duals()

        if self.debug:
            filename = os.path.join(
                helpers.extend_basic_path("lp_files"), "reegis.lp"
            )
            logging.info("Store lp-file in {0}.".format(filename))
            self.model.write(
                filename, io_options={"symbolic_solver_labels": True}
            )

        self.model.solve(
            solver=solver, solve_kwargs={"tee": tee, "logfile": logfile}
        )
        self.es.results["main"] = outputlib.processing.results(self.model)
        self.es.results["meta"] = outputlib.processing.meta_results(self.model)
        self.es.results["param"] = outputlib.processing.parameter_as_dict(
            self.es
        )
        self.es.results["meta"]["scenario"] = self.scenario_info(solver)
        self.es.results["meta"]["in_location"] = self.location
        self.es.results["meta"]["file_date"] = datetime.datetime.fromtimestamp(
            os.path.getmtime(self.location)
        )
        self.es.results["meta"]["oemof_version"] = logger.get_version()
        self.results = self.es.results["main"]

    def plot_nodes(self, show=None, filename=None, **kwargs):

        rm_nodes = kwargs.get("remove_nodes_with_substrings")

        g = graph.create_nx_graph(
            self.es, filename=filename, remove_nodes_with_substrings=rm_nodes
        )
        if show is True:
            draw_graph(g, **kwargs)
        return g


class Label(namedtuple("solph_label", ["cat", "tag", "subtag", "region"])):
    __slots__ = ()

    def __str__(self):
        return "_".join(map(str, self._asdict().values()))


class DeflexScenario(Scenario):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.extra_regions = kwargs.get("extra_regions", list())

    def create_nodes(self):
        # Create  a special dictionary that will raise an error if a key is
        # updated. This avoids the
        nodes = NodeDict()

        # Local volatile sources
        add_volatile_sources(self.table_collection, nodes)

        # Decentralised heating systems
        add_decentralised_heating_systems(
            self.table_collection, nodes, self.extra_regions
        )

        # Local electricity demand
        add_electricity_demand(self.table_collection, nodes)

        # Local district heating demand
        add_district_heating_systems(self.table_collection, nodes)

        # Connect electricity buses with transmission
        add_transmission_lines_between_electricity_nodes(
            self.table_collection, nodes
        )

        # Local power plants as Transformer and ExtractionTurbineCHP (chp)
        add_power_and_heat_plants(
            self.table_collection, nodes, self.extra_regions
        )

        # Storages
        if "storages" in self.table_collection:
            add_storages(self.table_collection, nodes)

        # Add shortage excess to every bus
        add_shortage_excess(nodes)
        return nodes


def create_fuel_bus_with_source(nodes, fuel, region, data):
    bus_label = Label("bus", "commodity", fuel.replace(" ", "_"), region)
    if bus_label not in nodes:
        nodes[bus_label] = solph.Bus(label=bus_label)

    cs_label = Label("source", "commodity", fuel.replace(" ", "_"), region)

    if cs_label not in nodes:
        nodes[cs_label] = solph.Source(
            label=cs_label,
            outputs={
                nodes[bus_label]: solph.Flow(
                    variable_costs=data.loc["costs", fuel.replace("_", " ")],
                    emission=data.loc["emission", fuel.replace("_", " ")],
                )
            },
        )


def add_volatile_sources(table_collection, nodes):
    """

    Parameters
    ----------
    table_collection
    nodes

    Returns
    -------

    """
    logging.debug("Add volatile sources to nodes dictionary.")
    vs = table_collection["volatile_source"]

    for region in vs.columns.get_level_values(0).unique():
        for vs_type in vs[region].columns:
            vs_label = Label("source", "ee", vs_type, region)
            capacity = vs.loc["capacity", (region, vs_type)]
            try:
                feedin = table_collection["volatile_series"][region, vs_type]
            except KeyError:
                if capacity > 0:
                    msg = "Missing time series for {0} (capacity: {1}) in {2}."
                    raise ValueError(msg.format(vs_type, capacity, region))
                feedin = [0]
            bus_label = Label("bus", "electricity", "all", region)
            if bus_label not in nodes:
                nodes[bus_label] = solph.Bus(label=bus_label)
            if capacity * sum(feedin) > 0:
                nodes[vs_label] = solph.Source(
                    label=vs_label,
                    outputs={
                        nodes[bus_label]: solph.Flow(
                            actual_value=feedin,
                            nominal_value=capacity,
                            fixed=True,
                            emission=0,
                        )
                    },
                )


def add_decentralised_heating_systems(table_collection, nodes, extra_regions):
    logging.debug("Add decentralised_heating_systems to nodes dictionary.")
    cs = table_collection["commodity_source"]["DE"]
    dts = table_collection["demand_series"]
    dh = table_collection["decentralised_heat"]
    demand_regions = list({"DE_demand"}.union(set(extra_regions)))

    for d_region in demand_regions:
        region_name = d_region.replace("_demand", "")

        if region_name not in dh:
            data_name = "DE_demand"
        else:
            data_name = d_region

        fuels = [f for f in dh[data_name].columns if f in dts[d_region]]
        for fuel in fuels:
            src = dh.loc["source", (data_name, fuel)]
            bus_label = Label(
                "bus", "commodity", src.replace(" ", "_"), region_name
            )

            # Check if source bus exists
            if bus_label not in nodes:
                create_fuel_bus_with_source(nodes, src, region_name, cs)

            # Create heating bus as Bus
            heat_bus_label = Label(
                "bus", "heat", fuel.replace(" ", "_"), region_name
            )
            nodes[heat_bus_label] = solph.Bus(label=heat_bus_label)

            # Create heating system as Transformer
            trsf_label = Label(
                "trsf", "heat", fuel.replace(" ", "_"), region_name
            )

            efficiency = float(dh.loc["efficiency", (data_name, fuel)])

            nodes[trsf_label] = solph.Transformer(
                label=trsf_label,
                inputs={nodes[bus_label]: solph.Flow()},
                outputs={nodes[heat_bus_label]: solph.Flow()},
                conversion_factors={nodes[heat_bus_label]: efficiency},
            )

            # Create demand as Sink
            d_heat_demand_label = Label(
                "demand", "heat", fuel.replace(" ", "_"), region_name
            )
            nodes[d_heat_demand_label] = solph.Sink(
                label=d_heat_demand_label,
                inputs={
                    nodes[heat_bus_label]: solph.Flow(
                        actual_value=dts[d_region, fuel],
                        nominal_value=1,
                        fixed=True,
                    )
                },
            )


def add_electricity_demand(table_collection, nodes):
    logging.debug("Add local electricity demand to nodes dictionary.")
    dts = table_collection["demand_series"]
    dts.columns = dts.columns.swaplevel()
    for region in dts["electrical_load"].columns:
        if dts["electrical_load"][region].sum() > 0:
            bus_label = Label("bus", "electricity", "all", region)
            if bus_label not in nodes:
                nodes[bus_label] = solph.Bus(label=bus_label)
            elec_demand_label = Label("demand", "electricity", "all", region)
            nodes[elec_demand_label] = solph.Sink(
                label=elec_demand_label,
                inputs={
                    nodes[bus_label]: solph.Flow(
                        actual_value=dts["electrical_load", region],
                        nominal_value=1,
                        fixed=True,
                    )
                },
            )


def add_district_heating_systems(table_collection, nodes):
    logging.debug("Add district heating systems to nodes dictionary.")
    dts = table_collection["demand_series"]
    for region in dts["district heating"].columns:
        if dts["district heating"][region].sum() > 0:
            bus_label = Label("bus", "heat", "district", region)
            if bus_label not in nodes:
                nodes[bus_label] = solph.Bus(label=bus_label)
            heat_demand_label = Label("demand", "heat", "district", region)
            nodes[heat_demand_label] = solph.Sink(
                label=heat_demand_label,
                inputs={
                    nodes[bus_label]: solph.Flow(
                        actual_value=dts["district heating", region],
                        nominal_value=1,
                        fixed=True,
                    )
                },
            )


def add_transmission_lines_between_electricity_nodes(table_collection, nodes):
    logging.debug("Add transmission lines to nodes dictionary.")
    power_lines = table_collection["transmission"]["electrical"]
    for idx, values in power_lines.iterrows():
        b1, b2 = idx.split("-")
        lines = [(b1, b2), (b2, b1)]
        for line in lines:
            line_label = Label("line", "electricity", line[0], line[1])
            bus_label_in = Label("bus", "electricity", "all", line[0])
            bus_label_out = Label("bus", "electricity", "all", line[1])
            if bus_label_in not in nodes:
                raise ValueError(
                    "Bus {0} missing for power line from {0} to {1}".format(
                        bus_label_in, bus_label_out
                    )
                )
            if bus_label_out not in nodes:
                raise ValueError(
                    "Bus {0} missing for power line from {0} to {1}".format(
                        bus_label_out, bus_label_in
                    )
                )
            if values.capacity != float("inf"):
                logging.debug(
                    "Line {0} has a capacity of {1}".format(
                        line_label, values.capacity
                    )
                )
                nodes[line_label] = solph.Transformer(
                    label=line_label,
                    inputs={nodes[bus_label_in]: solph.Flow()},
                    outputs={
                        nodes[bus_label_out]: solph.Flow(
                            nominal_value=values.capacity
                        )
                    },
                    conversion_factors={
                        nodes[bus_label_out]: values.efficiency
                    },
                )
            else:
                logging.debug(
                    "Line {0} has no capacity limit".format(line_label)
                )
                nodes[line_label] = solph.Transformer(
                    label=line_label,
                    inputs={nodes[bus_label_in]: solph.Flow()},
                    outputs={nodes[bus_label_out]: solph.Flow()},
                    conversion_factors={
                        nodes[bus_label_out]: values.efficiency
                    },
                )


def add_power_and_heat_plants(table_collection, nodes, extra_regions):
    trsf = table_collection["transformer"]
    cs = table_collection["commodity_source"]["DE"]

    for region in trsf.columns.get_level_values(0).unique():
        bus_heat = Label("bus", "heat", "district", region)
        bus_elec = Label("bus", "electricity", "all", region)
        for fuel in trsf[region].columns:
            # Connect to global fuel bus if not defined as extra region
            if region in extra_regions:
                bus_fuel = Label(
                    "bus", "commodity", fuel.replace(" ", "_"), region
                )
                create_fuel_bus_with_source(nodes, fuel, region, cs)
            else:
                bus_fuel = Label(
                    "bus", "commodity", fuel.replace(" ", "_"), "DE"
                )
                if bus_fuel not in nodes:
                    create_fuel_bus_with_source(
                        nodes, fuel.replace(" ", "_"), "DE", cs
                    )
            params = trsf[region, fuel]

            # Create power plants as 1x1 Transformer
            if params.capacity > 0:
                # Define output flow with or without summed_max attribute
                if params.limit_elec_pp == float("inf"):
                    outflow = solph.Flow(nominal_value=params.capacity)
                else:
                    smax = params.limit_elec_pp / params.capacity
                    outflow = solph.Flow(
                        nominal_value=params.capacity, summed_max=smax
                    )

                trsf_label = Label(
                    "trsf", "pp", fuel.replace(" ", "_"), region
                )
                nodes[trsf_label] = solph.Transformer(
                    label=trsf_label,
                    inputs={nodes[bus_fuel]: solph.Flow()},
                    outputs={nodes[bus_elec]: outflow},
                    conversion_factors={nodes[bus_elec]: params.efficiency},
                )

            # Create chp plants as 1x2 Transformer
            if params.capacity_heat_chp > 0:
                trsf_label = Label(
                    "trsf", "chp", fuel.replace(" ", "_"), region
                )

                smax = (params.limit_heat_chp / params.efficiency_heat_chp) / (
                    params.capacity_heat_chp / params.efficiency_heat_chp
                )

                nodes[trsf_label] = solph.Transformer(
                    label=trsf_label,
                    inputs={
                        nodes[bus_fuel]: solph.Flow(
                            nominal_value=(
                                params.capacity_heat_chp
                                / params.efficiency_heat_chp
                            ),
                            summed_max=smax,
                        )
                    },
                    outputs={
                        nodes[bus_elec]: solph.Flow(),
                        nodes[bus_heat]: solph.Flow(),
                    },
                    conversion_factors={
                        nodes[bus_elec]: params.efficiency_elec_chp,
                        nodes[bus_heat]: params.efficiency_heat_chp,
                    },
                )

            # Create heat plants as 1x1 Transformer
            if params.capacity_hp > 0:
                trsf_label = Label(
                    "trsf", "hp", fuel.replace(" ", "_"), region
                )
                smax = params.limit_hp / params.capacity_hp

                nodes[trsf_label] = solph.Transformer(
                    label=trsf_label,
                    inputs={nodes[bus_fuel]: solph.Flow()},
                    outputs={
                        nodes[bus_heat]: solph.Flow(
                            nominal_value=params.capacity_hp, summed_max=smax
                        )
                    },
                    conversion_factors={nodes[bus_heat]: params.efficiency_hp},
                )


def add_storages(table_collection, nodes):
    storages = table_collection["storages"]
    storages.columns = storages.columns.swaplevel()
    for region in storages["phes"].columns:
        storage_label = Label("storage", "electricity", "phes", region)
        bus_label = Label("bus", "electricity", "all", region)
        params = storages["phes"][region]
        nodes[storage_label] = solph.components.GenericStorage(
            label=storage_label,
            inputs={nodes[bus_label]: solph.Flow(nominal_value=params.pump)},
            outputs={
                nodes[bus_label]: solph.Flow(nominal_value=params.turbine)
            },
            nominal_storage_capacity=params.energy,
            loss_rate=0,
            initial_storage_level=None,
            inflow_conversion_factor=params.pump_eff,
            outflow_conversion_factor=params.turbine_eff,
        )


def add_shortage_excess(nodes):
    bus_keys = [key for key in nodes.keys() if "bus" in key.cat]
    for key in bus_keys:
        excess_label = Label("excess", key.tag, key.subtag, key.region)
        nodes[excess_label] = solph.Sink(
            label=excess_label, inputs={nodes[key]: solph.Flow()}
        )
        shortage_label = Label("shortage", key.tag, key.subtag, key.region)
        nodes[shortage_label] = solph.Source(
            label=shortage_label,
            outputs={nodes[key]: solph.Flow(variable_costs=900)},
        )


def draw_graph(
    grph,
    edge_labels=True,
    node_color="#AFAFAF",
    edge_color="#CFCFCF",
    plot=True,
    node_size=2000,
    with_labels=True,
    arrows=True,
    layout="neato",
):
    """
    Draw a graph. This function will be removed in future versions.

    Parameters
    ----------
    grph : networkxGraph
        A graph to draw.
    edge_labels : boolean
        Use nominal values of flow as edge label
    node_color : dict or string
        Hex color code oder matplotlib color for each node. If string, all
        colors are the same.

    edge_color : string
        Hex color code oder matplotlib color for edge color.

    plot : boolean
        Show matplotlib plot.

    node_size : integer
        Size of nodes.

    with_labels : boolean
        Draw node labels.

    arrows : boolean
        Draw arrows on directed edges. Works only if an optimization_model has
        been passed.
    layout : string
        networkx graph layout, one of: neato, dot, twopi, circo, fdp, sfdp.
    """
    if type(node_color) is dict:
        node_color = [node_color.get(g, "#AFAFAF") for g in grph.nodes()]

    # set drawing options
    options = {
        "prog": "dot",
        "with_labels": with_labels,
        "node_color": node_color,
        "edge_color": edge_color,
        "node_size": node_size,
        "arrows": arrows,
    }

    # draw graph
    pos = nx.drawing.nx_agraph.graphviz_layout(grph, prog=layout)

    nx.draw(grph, pos=pos, **options)

    # add edge labels for all edges
    if edge_labels is True and plt:
        labels = nx.get_edge_attributes(grph, "weight")
        nx.draw_networkx_edge_labels(grph, pos=pos, edge_labels=labels)

    # show output
    if plot is True:
        plt.show()


if __name__ == "__main__":
    pass



#! /usr/bin/env python

from setuptools import setup, find_packages
import os
import deflex


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="deflex",
    version=deflex.__version__,
    author="Uwe Krien",
    author_email="krien@uni-bremen.de",
    description=(
        "A multi-sectoral model of Germany with flexible multi regions."
    ),
    namespace_package=["deflex"],
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    packages=find_packages(),
    package_dir={"deflex": "deflex"},
    url="https://github.com/reegis/deflex",
    license="MIT",
    extras_require={"dev": ["nose", "sphinx", "sphinx_rtd_theme", "requests"]},
    install_requires=[
        "oemof >= 0.3.0",
        "pandas >= 0.17.0",
        "reegis > v0.1.1",
        "demandlib",
        "workalendar",
        "networkx",
        "numpy",
        "rtree",
        "xlrd",
        "xlwt",
        "dill",
        "matplotlib",
    ],
    package_data={
        "deflex": [
            os.path.join("data", "static", "*.csv"),
            os.path.join("data", "static", "*.txt"),
            os.path.join("data", "geometries", "*.csv"),
            os.path.join("data", "geometries", "*.geojson"),
            "*.ini",
        ]
    },
)



# -*- coding: utf-8 -*-

"""Create a basic scenario from the internal data structure.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"

# Python libraries
import os
import logging
import calendar
from collections import namedtuple

# External libraries
import pandas as pd

# internal modules
from reegis import commodity_sources
from reegis import bmwi
from reegis import storages
from reegis import coastdat
from reegis import demand_elec
from reegis import energy_balance
from reegis import powerplants as reegis_powerplants
from deflex import powerplants
from deflex import demand
from deflex import transmission
from deflex import scenario_tools
from deflex import geometries
from deflex import config as cfg


def create_scenario(
    regions, year, name, round_values=0, weather_year=None, copperplate=False
):
    table_collection = {}

    logging.info("BASIC SCENARIO - STORAGES")
    table_collection["storages"] = scenario_storages(regions, year, name)

    logging.info("BASIC SCENARIO - POWER PLANTS")
    table_collection = scenario_powerplants(
        table_collection, regions, year, name, round_values
    )

    logging.info("BASIC SCENARIO - TRANSMISSION")
    table_collection["transmission"] = scenario_transmission(
        table_collection, regions, name, copperplate=copperplate
    )

    logging.info("BASIC SCENARIO - CHP PLANTS")
    table_collection = scenario_chp(
        table_collection, regions, year, name, weather_year=weather_year
    )

    logging.info("BASIC SCENARIO - DECENTRALISED HEAT")
    table_collection["decentralised_heat"] = scenario_decentralised_heat()

    logging.info("BASIC SCENARIO - SOURCES")
    table_collection = scenario_commodity_sources(table_collection, year)
    table_collection["volatile_series"] = scenario_feedin(
        regions, year, name, weather_year=weather_year
    )

    logging.info("BASIC SCENARIO - DEMAND")
    table_collection["demand_series"] = scenario_demand(
        regions, year, name, weather_year=weather_year
    )
    return table_collection


def scenario_storages(regions, year, name):
    """
    Fetch storage, pump and turbine capacity and their efficiency of
    hydroelectric storages for each deflex region.

    Parameters
    ----------
    regions
    year
    name

    Returns
    -------

    Examples
    --------
    >>> regions=geometries.deflex_regions(rmap='de17')
    >>> deflex_storages=scenario_storages(regions, 2012, 'de17')
    >>> list(deflex_storages.columns.get_level_values(0))
    ['DE01', 'DE03', 'DE05', 'DE06', 'DE08', 'DE09', 'DE14', 'DE15', 'DE16']
    >>> int(deflex_storages.loc['turbine', 'DE03'])
    220
    >>> int(deflex_storages.loc['energy', 'DE16'])
    12115
    """
    stor = storages.pumped_hydroelectric_storage_by_region(
        regions, year, name
    ).transpose()
    return pd.concat([stor], axis=1, keys=["phes"]).swaplevel(0, 1, 1)


def scenario_powerplants(table_collection, regions, year, name, round_values):
    """Get power plants for the scenario year

    Examples
    --------
    >>> regions=geometries.deflex_regions(rmap='de21')
    >>> pp=scenario_powerplants(
    ...     dict(), regions, 2014, 'de21', 1)  # doctest: +SKIP
    >>> float(pp['volatile_source']['DE03', 'wind'])  # doctest: +SKIP
    3052.8
    >>> float(pp['transformer'].loc[
    ...     'capacity', ('DE03', 'lignite')])  # doctest: +SKIP
    1135.6
    """
    pp = powerplants.get_deflex_pp_by_year(
        regions, year, name, overwrite_capacity=True
    )
    return create_powerplants(pp, table_collection, year, name, round_values)


def create_powerplants(
    pp,
    table_collection,
    year,
    region_column="deflex_region",
    round_values=None,
):
    """This function works for all power plant tables with an equivalent
    structure e.g. power plants by state or other regions."""
    logging.info("Adding power plants to your scenario.")

    replace_names = cfg.get_dict("source_names")
    replace_names.update(cfg.get_dict("source_groups"))

    pp["energy_source_level_2"].replace(replace_names, inplace=True)

    pp["model_classes"] = pp["energy_source_level_2"].replace(
        cfg.get_dict("model_classes")
    )

    pp = pp.groupby(
        ["model_classes", region_column, "energy_source_level_2"]
    ).sum()[["capacity", "capacity_in"]]

    for model_class in pp.index.get_level_values(level=0).unique():
        pp_class = pp.loc[model_class]
        if model_class != "volatile_source":
            pp_class["efficiency"] = (
                pp_class["capacity"] / pp_class["capacity_in"] * 100
            )
        del pp_class["capacity_in"]
        if round_values is not None:
            pp_class = pp_class.round(round_values)
        if "efficiency" in pp_class:
            pp_class["efficiency"] = pp_class["efficiency"].div(100)
        pp_class = pp_class.transpose()
        pp_class.index.name = "parameter"
        table_collection[model_class] = pp_class
    table_collection = add_pp_limit(table_collection, year)
    return table_collection


def add_pp_limit(table_collection, year):
    """

    Parameters
    ----------
    table_collection
    year

    Returns
    -------

    """
    if len(cfg.get_dict("limited_transformer").keys()) > 0:
        # Multiply with 1000 to get MWh (bmwi: GWh)
        repp = bmwi.bmwi_re_energy_capacity() * 1000
        trsf = table_collection["transformer"]
        for limit_trsf in cfg.get_dict("limited_transformer").keys():
            trsf = table_collection["transformer"]
            try:
                limit = repp.loc[year, (limit_trsf, "energy")]
            except KeyError:
                msg = "Cannot calculate limit for {0} in {1}."
                raise ValueError(msg.format(limit_trsf, year))
            cap_sum = trsf.loc[
                "capacity", (slice(None), slice(limit_trsf))
            ].sum()
            for region in trsf.columns.get_level_values(level=0).unique():
                trsf.loc["limit_elec_pp", (region, limit_trsf)] = round(
                    trsf.loc["capacity", (region, limit_trsf)]
                    / cap_sum
                    * limit
                    + 0.5
                )

        trsf.loc["limit_elec_pp"] = trsf.loc["limit_elec_pp"].fillna(
            float("inf")
        )

        table_collection["transformer"] = trsf
    return table_collection


def scenario_transmission(table_collection, regions, name, copperplate=False):
    """Get power plants for the scenario year

    Examples
    --------
    >>> regions=geometries.deflex_regions(rmap='de21')  # doctest: +SKIP
    >>> pp=scenario_powerplants(dict(), regions, 2014, 'de21', 1
    ...     )  # doctest: +SKIP
    >>> lines=scenario_transmission(pp, regions, 'de21')  # doctest: +SKIP
    >>> int(lines.loc['DE07-DE05', ('electrical', 'capacity')]
    ...     )  # doctest: +SKIP
    1978
    >>> int(lines.loc['DE07-DE05', ('electrical', 'distance')]
    ...     )  # doctest: +SKIP
    199
    >>> float(lines.loc['DE07-DE05', ('electrical', 'efficiency')]
    ...     )  # doctest: +SKIP
    0.9
    >>> lines=scenario_transmission(pp, regions, 'de21', copperplate=True
    ...     )  # doctest: +SKIP
    >>> float(lines.loc['DE07-DE05', ('electrical', 'capacity')]
    ...     )  # doctest: +SKIP
    inf
    >>> float(lines.loc['DE07-DE05', ('electrical', 'distance')]
    ...     )  # doctest: +SKIP
    nan
    >>> float(lines.loc['DE07-DE05', ('electrical', 'efficiency')]
    ...     )  # doctest: +SKIP
    1.0
    """
    vs = table_collection["volatile_source"]

    # This should be done automatic e.g. if representative point outside the
    # landmass polygon.
    offshore_regions = geometries.divide_off_and_onshore(regions).offshore

    if name in ["de21", "de22"] and not copperplate:
        elec_trans = transmission.get_electrical_transmission_renpass()
        general_efficiency = cfg.get("transmission", "general_efficiency")
        if general_efficiency is not None:
            elec_trans["efficiency"] = general_efficiency
        else:
            msg = (
                "The calculation of the efficiency by distance is not yet "
                "implemented"
            )
            raise NotImplementedError(msg)
    else:
        elec_trans = transmission.get_electrical_transmission_default()

    # Set transmission capacity of offshore power lines to installed capacity
    # Multiply the installed capacity with 1.1 to get a buffer of 10%.
    for offreg in offshore_regions:
        elec_trans.loc[elec_trans.index.str.contains(offreg), "capacity"] = (
            vs[offreg].sum().sum() * 1.1
        )

    elec_trans = pd.concat(
        [elec_trans], axis=1, keys=["electrical"]
    ).sort_index(1)
    if cfg.get("init", "map") == "de22" and not copperplate:
        elec_trans.loc["DE22-DE01", ("electrical", "efficiency")] = 0.9999
        elec_trans.loc["DE22-DE01", ("electrical", "capacity")] = 9999999
    return elec_trans


def scenario_commodity_sources(table_collection, year):
    """

    Parameters
    ----------
    year
    table_collection

    Returns
    -------

    Examples
    --------
    >>> regions=geometries.deflex_regions(rmap='de21')  # doctest: +SKIP
    >>> pp=scenario_powerplants(dict(), regions, 2014, 'de21', 1
    ...     )  # doctest: +SKIP
    >>> src=scenario_commodity_sources(pp, 2014)  # doctest: +SKIP
    >>> src=src['commodity_source']  # doctest: +SKIP
    >>> round(src.loc['costs', ('DE', 'hard coal')], 2)  # doctest: +SKIP
    8.93
    >>> round(src.loc['emission',  ('DE', 'natural gas')], 2)  # doctest: +SKIP
    201.24
    """
    commodity_src = create_commodity_sources(year)
    commodity_src = commodity_src.swaplevel().unstack()

    msg = (
        "The unit for {0} of the source is '{1}'. "
        "Will multiply it with {2} to get '{3}'."
    )

    converter = {
        "costs": ["costs", "EUR/J", 1e9 * 3.6, "EUR/MWh"],
        "emission": ["emission", "g/J", 1e6 * 3.6, "kg/MWh"],
    }

    transformer_list = (
        table_collection["transformer"]
        .columns.get_level_values(level=1)
        .unique()
    )

    # Delete unused sources
    for col in commodity_src.columns:
        if col not in transformer_list:
            del commodity_src[col]

    # convert units
    for key in converter.keys():
        commodity_src.loc[key] = commodity_src.loc[key].multiply(
            converter[key][2]
        )
        logging.warning(msg.format(*converter[key]))

    # Add region level to be consistent to other tables
    commodity_src.columns = pd.MultiIndex.from_product(
        [["DE"], commodity_src.columns]
    )

    table_collection["commodity_source"] = commodity_src
    return table_collection


def create_commodity_sources(year, use_znes_2014=True):
    """

    Parameters
    ----------
    year
    use_znes_2014

    Returns
    -------

    """
    cs = commodity_sources.get_commodity_sources()
    rename_cols = {
        key.lower(): value
        for key, value in cfg.get_dict("source_names").items()
    }
    cs = cs.rename(columns=rename_cols)
    cs_year = cs.loc[year]
    if use_znes_2014:
        before = len(cs_year[cs_year.isnull()])
        cs_year = cs_year.fillna(cs.loc[2014])
        after = len(cs_year[cs_year.isnull()])
        if before - after > 0:
            logging.warning("Values were replaced with znes2014 data.")
    cs_year.sort_index(inplace=True)
    return cs_year


def scenario_demand(regions, year, name, weather_year=None):
    """

    Parameters
    ----------
    regions
    year
    name
    weather_year

    Returns
    -------

    Examples
    --------
    >>> regions=geometries.deflex_regions(rmap='de21')  # doctest: +SKIP
    >>> my_demand=scenario_demand(regions, 2014, 'de21')  # doctest: +SKIP
    >>> int(my_demand['DE01', 'district heating'].sum())  # doctest: +SKIP
    18639262
    >>> int(my_demand['DE05', 'electrical_load'].sum())  # doctest: +SKIP
    10069

    """
    demand_series = scenario_elec_demand(
        pd.DataFrame(), regions, year, name, weather_year=weather_year
    )
    demand_series = scenario_heat_demand(
        demand_series, regions, year, weather_year=weather_year
    )
    return demand_series


def scenario_heat_demand(table, regions, year, weather_year=None):
    idx = table.index  # Use the index of the existing time series
    table = pd.concat(
        [
            table,
            demand.get_heat_profiles_deflex(
                regions, year, idx, weather_year=weather_year
            ),
        ],
        axis=1,
    )
    return table.sort_index(1)


def scenario_elec_demand(table, regions, year, name, weather_year=None):
    if weather_year is None:
        demand_year = year
    else:
        demand_year = weather_year

    df = demand_elec.get_entsoe_profile_by_region(
        regions, demand_year, name, annual_demand="bmwi"
    )
    df = pd.concat([df], axis=1, keys=["electrical_load"]).swaplevel(0, 1, 1)
    df = df.reset_index(drop=True)
    if not calendar.isleap(year) and len(df) > 8760:
        df = df.iloc[:8760]
    return pd.concat([table, df], axis=1).sort_index(1)


def scenario_feedin(regions, year, name, weather_year=None):
    """

    Parameters
    ----------
    regions
    year
    name
    weather_year

    Returns
    -------

    Examples
    --------
    >>> regions=geometries.deflex_regions(rmap='de21')  # doctest: +SKIP
    >>> f=scenario_feedin(regions, 2014, 'de21')  # doctest: +SKIP
    >>> f['DE01'].sum()  # doctest: +SKIP
    geothermal    4380.000000
    hydro         1346.632529
    solar          913.652083
    wind          2152.983589
    dtype: float64
    >>> f['DE16'].sum()  # doctest: +SKIP
    geothermal    4380.000000
    hydro         1346.632529
    solar          903.527200
    wind          1753.673492
    dtype: float64
    """
    wy = weather_year
    try:
        feedin = coastdat.scenario_feedin(year, name, weather_year=wy)
    except FileNotFoundError:
        coastdat.get_feedin_per_region(year, regions, name, weather_year=wy)
        feedin = coastdat.scenario_feedin(year, name, weather_year=wy)
    return feedin


def scenario_decentralised_heat():
    filename = os.path.join(
        cfg.get("paths", "data_deflex"), cfg.get("heating", "table")
    )
    return pd.read_csv(filename, header=[0, 1], index_col=[0])


def scenario_chp(table_collection, regions, year, name, weather_year=None):
    """

    Parameters
    ----------
    table_collection
    regions
    year
    name
    weather_year

    Returns
    -------

    Examples
    --------
    >>> regions=geometries.deflex_regions(rmap='de21')  # doctest: +SKIP
    >>> pp=scenario_powerplants(dict(), regions, 2014, 'de21', 1
    ...     )  # doctest: +SKIP
    >>> int(pp['transformer'].loc['capacity', ('DE01', 'hard coal')]
    ...     )  # doctest: +SKIP
    1291
    >>> transf=scenario_chp(pp, regions, 2014, 'de21')  # doctest: +SKIP
    >>> transf=transf['transformer']  # doctest: +SKIP
    >>> int(transf.loc['capacity', ('DE01', 'hard coal')])  # doctest: +SKIP
    485
    >>> int(transf.loc['capacity_elec_chp', ('DE01', 'hard coal')]
    ...     )  # doctest: +SKIP
    806
    """
    # values from heat balance

    cb = energy_balance.get_transformation_balance_by_region(
        regions, year, name
    )
    cb.rename(columns={"re": cfg.get("chp", "renewable_source")}, inplace=True)
    heat_b = reegis_powerplants.calculate_chp_share_and_efficiency(cb)

    heat_demand = demand.get_heat_profiles_deflex(
        regions, year, weather_year=weather_year
    )
    return chp_table(heat_b, heat_demand, table_collection)


def chp_table(heat_b, heat_demand, table_collection, regions=None):
    trsf = table_collection["transformer"]
    trsf = trsf.fillna(0)

    rows = ["Heizkraftwerke der allgemeinen Versorgung (nur KWK)", "Heizwerke"]
    if regions is None:
        regions = sorted(heat_b.keys())

    logging.info("start")
    for region in regions:
        eta_hp = round(heat_b[region]["sys_heat"] * heat_b[region]["hp"], 2)
        eta_heat_chp = round(
            heat_b[region]["sys_heat"] * heat_b[region]["heat_chp"], 2
        )
        eta_elec_chp = round(heat_b[region]["elec_chp"], 2)

        # Due to the different efficiency between heat from chp-plants and
        # heat from heat-plants the share of the output is different to the
        # share of the input. As heat-plants will produce more heat per fuel
        # factor will be greater than 1 and for chp-plants smaller than 1.
        out_share_factor_chp = heat_b[region]["out_share_factor_chp"]
        out_share_factor_hp = heat_b[region]["out_share_factor_hp"]

        # Remove 'district heating' and 'electricity' and spread the share
        # to the remaining columns.
        share = pd.DataFrame(columns=heat_b[region]["fuel_share"].columns)
        for row in rows:
            tmp = heat_b[region]["fuel_share"].loc[region, :, row]
            tot = float(tmp["total"])

            d = float(tmp["district heating"] + tmp["electricity"])
            tmp = tmp + tmp / (tot - d) * d
            tmp = tmp.reset_index(drop=True)
            share.loc[row] = tmp.loc[0]
        del share["district heating"]
        del share["electricity"]

        # Remove the total share
        del share["total"]

        max_val = float(heat_demand[region]["district heating"].max())
        sum_val = float(heat_demand[region]["district heating"].sum())

        for fuel in share.columns:
            if fuel == "gas":
                src = "natural gas"
            else:
                src = fuel

            # CHP
            trsf.loc["limit_heat_chp", (region, src)] = round(
                sum_val * share.loc[rows[0], fuel] * out_share_factor_chp + 0.5
            )
            cap_heat_chp = round(
                max_val * share.loc[rows[0], fuel] * out_share_factor_chp
                + 0.005,
                2,
            )
            trsf.loc["capacity_heat_chp", (region, src)] = cap_heat_chp
            cap_elec = cap_heat_chp / eta_heat_chp * eta_elec_chp
            trsf.loc["capacity_elec_chp", (region, src)] = round(cap_elec, 2)
            trsf[region] = trsf[region].fillna(0)
            trsf.loc["capacity", (region, src)] = round(
                trsf.loc["capacity", (region, src)] - cap_elec
            )

            # If the power plant limit is not 'inf' the limited electricity
            # output of the chp plant has to be subtracted from the power plant
            # limit because this is related to the overall electricity output.
            if not trsf.loc["limit_elec_pp", (region, src)] == float("inf"):
                trsf.loc["limit_elec_pp", (region, src)] -= round(
                    trsf.loc["limit_heat_chp", (region, src)]
                    / eta_heat_chp
                    * eta_elec_chp
                )

            # HP
            trsf.loc["limit_hp", (region, src)] = round(
                sum_val * share.loc[rows[1], fuel] * out_share_factor_hp + 0.5
            )
            trsf.loc["capacity_hp", (region, src)] = round(
                max_val * share.loc[rows[1], fuel] * out_share_factor_hp
                + 0.005,
                2,
            )
            if trsf.loc["capacity_hp", (region, src)] > 0:
                trsf.loc["efficiency_hp", (region, src)] = eta_hp
            if cap_heat_chp * cap_elec > 0:
                trsf.loc["efficiency_heat_chp", (region, src)] = eta_heat_chp
                trsf.loc["efficiency_elec_chp", (region, src)] = eta_elec_chp

    logging.info("Done")

    trsf.sort_index(axis=1, inplace=True)
    for col in trsf.sum().loc[trsf.sum() == 0].index:
        del trsf[col]
    trsf[trsf < 0] = 0

    table_collection["transformer"] = trsf
    return table_collection


def clean_time_series(table_collection):
    dts = table_collection["demand_series"]
    vts = table_collection["volatile_series"]
    vs = table_collection["volatile_source"]

    regions = list(dts.columns.get_level_values(0).unique())
    regions.remove("DE_demand")
    for reg in regions:
        for load in ["district heating", "electrical_load"]:
            if dts[reg].get(load) is not None:
                if dts[reg, load].sum() == 0:
                    msg = (
                        "Removing {0} time series of region {1} because"
                        "sum of time series is {2}"
                    )
                    logging.debug(msg.format(load, reg, dts[reg, load].sum()))
                    del dts[reg, load]

    regions = list(vts.columns.get_level_values(0).unique())
    for reg in regions:
        for t in ["hydro", "solar", "wind", "geothermal"]:
            # if the column does not exist or is 0 the corresponding column
            # of the time_series table can be removed.
            if vs[reg].get(t) is None or vs[reg].get(t).sum() == 0:
                if vts.get(reg) is not None:
                    if vts[reg].get(t) is not None:
                        msg = (
                            "Removing {0} time series of region {1} "
                            "because installed capacity is {2}"
                        )
                        logging.debug(msg.format(t, reg, vs[reg].get(t)))
                        del vts[reg, t]

    return table_collection


def create_basic_scenario(
    year,
    rmap=None,
    path=None,
    csv_dir=None,
    xls_name=None,
    round_values=None,
    only_out=None,
):
    """
    Create a basic scenario for a given year and region-set.

    Parameters
    ----------
    year : int
    rmap : str
    path : str
    csv_dir : str
    xls_name : str
    round_values : bool
    only_out : str

    Returns
    -------
    namedtuple : Path

    Examples
    --------
    >>> year=2014  # doctest: +SKIP
    >>> my_rmap='de21'  # doctest: +SKIP
    >>> p=create_basic_scenario(year, rmap=my_rmap)  # doctest: +SKIP
    >>> print("Xls path: {0}".format(p.xls))  # doctest: +SKIP
    >>> print("Csv path: {0}".format(p.csv))  # doctest: +SKIP

    """
    paths = namedtuple("paths", "xls, csv")
    if rmap is not None:
        cfg.tmp_set("init", "map", rmap)
    name = cfg.get("init", "map")
    regions = geometries.deflex_regions(rmap=cfg.get("init", "map"))

    table_collection = create_scenario(regions, year, name, round_values)
    table_collection = clean_time_series(table_collection)
    name = "{0}_{1}_{2}".format("deflex", year, cfg.get("init", "map"))
    sce = scenario_tools.Scenario(
        table_collection=table_collection, name=name, year=year
    )

    if path is None:
        path = os.path.join(cfg.get("paths", "scenario"), "deflex", str(year))

    if only_out == "xls":
        csv_path = None
    elif csv_dir is None:
        csv_path = os.path.join(path, "{0}_csv".format(name))
    else:
        csv_path = os.path.join(path, csv_dir)

    if only_out == "csv":
        xls_path = None
    elif xls_name is None:
        xls_path = os.path.join(path, name + ".xls")
    else:
        xls_path = os.path.join(path, xls_name)

    fullpath = paths(xls=xls_path, csv=csv_path)
    if not only_out == "csv":
        os.makedirs(path, exist_ok=True)
        sce.to_excel(fullpath.xls)
    if not only_out == "xls":
        os.makedirs(csv_path, exist_ok=True)
        sce.to_csv(fullpath.csv)

    return fullpath


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""
Reegis config reader.

SPDX-FileCopyrightText: 2016-2019 Uwe Krien <krien@uni-bremen.de>

SPDX-License-Identifier: MIT
"""
__copyright__ = "Uwe Krien <krien@uni-bremen.de>"
__license__ = "MIT"


# Python libraries
import os
from reegis.config import *


_loaded = False
if not _loaded:
    init(paths=[os.path.dirname(__file__)])
    _loaded = True
